import json
from datetime import timedelta

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.db.utils import IntegrityError
from django.utils import timezone
from django.utils.module_loading import import_string

from .config import settings


class VerificationTokenManager(models.Manager):

    def deactivate(self, obj, slug=None, key=None):
        self.filter_active_tokens(obj, slug, key).update(is_active=False)

    def deactivate_and_create(self, obj, slug=None, extra_data=None, deactivate_old_tokens=True,
                              key_generator_kwargs=None, **kwargs):
        if deactivate_old_tokens:
            self.deactivate(obj, slug)

        return self._create(obj, slug=slug, extra_data=extra_data, key_generator_kwargs=key_generator_kwargs, **kwargs)

    def get_active_or_create(self, obj, slug=None, extra_data=None, key=None, key_generator_kwargs=None, **kwargs):
        token = self.filter_active_tokens(obj, slug, key).order_by('created_at').last()

        if token and token.is_valid:
            return token
        else:
            return self._create(obj, slug=slug, extra_data=extra_data, key_generator_kwargs=key_generator_kwargs,
                                **kwargs
            )

    def _create(self, obj, slug=None, extra_data=None, key_generator_kwargs=None, **kwargs):
        expiration_in_minutes = kwargs.pop('expiration_in_minutes', settings.DEFAULT_EXPIRATION)
        key_generator_kwargs = {} if key_generator_kwargs is None else key_generator_kwargs

        token = self.model(
            content_type=ContentType.objects.get_for_model(obj.__class__),
            object_id=obj.pk,
            slug=slug,
            expires_at=(timezone.now() + timedelta(minutes=expiration_in_minutes)) if expiration_in_minutes else None,
            key=self.model.generate_key(**key_generator_kwargs),
        )
        if extra_data:
            token.set_extra_data(extra_data)

        token.save()
        return token

    def exists_valid(self, obj, key, slug=None):
        for token in self.filter_active_tokens(obj, slug):
            if token.check_key(key):
                return True
        return False

    def filter_active_tokens(self, obj_or_class, slug=None, key=None):
        qs = self.filter(
            is_active=True,
            slug=slug,
            content_type=ContentType.objects.get_for_model(obj_or_class),
        )
        if isinstance(obj_or_class, models.Model):
            qs = qs.filter(object_id=obj_or_class.pk)
        return qs.filter(key=key) if key else qs


class VerificationToken(models.Model):
    """
    Specific verification tokens that can be send via e-mail to check user authorization (example password reset)
    """

    created_at = models.DateTimeField(auto_now_add=True, null=False, blank=False)
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.TextField(db_index=True)
    content_object = GenericForeignKey('content_type', 'object_id')
    key = models.CharField(null=False, blank=False, max_length=100, unique=True)
    expires_at = models.DateTimeField(null=True, blank=True, default=None)
    slug = models.SlugField(null=True, blank=True)
    is_active = models.BooleanField(null=False, blank=False, default=True)
    extra_data = models.TextField(null=True, blank=True)

    objects = VerificationTokenManager()

    @classmethod
    def generate_key(cls, generator=None, *args, **kwargs):
        """
        Generate random unique token key.
        """
        generator = settings.DEFAULT_KEY_GENERATOR if generator is None else generator
        generator_func = import_string(generator) if isinstance(generator, str) else generator

        key = generator_func(*args, **kwargs)
        try_generator_iterations = 1
        while cls.objects.filter(key=key).exists():
            if try_generator_iterations >= settings.MAX_RANDOM_KEY_ITERATIONS:
                raise IntegrityError('Could not produce unique key for verification token')
            try_generator_iterations += 1
            key = generator_func(*args, **kwargs)
        return key

    @property
    def is_valid(self):
        return (
            self.is_active and self.key and (self.expires_at is None or timezone.now() <= self.expires_at)
        )

    def check_key(self, key):
        """
        Returns True if verification key is correct and not expired
        """
        return self.is_valid and self.key == key

    def set_extra_data(self, extra_data):
        self.extra_data = json.dumps(extra_data)

    def get_extra_data(self):
        return json.loads(self.extra_data) if self.extra_data is not None else None

    def save(self, *args, **kwargs):
        if not self.key:
            self.key = self.generate_key()
        super().save(*args, **kwargs)

    def __str__(self):
        return self.key

    class Meta:
        ordering = ('-created_at',)



from datetime import timedelta
from io import StringIO

from django.core.management import call_command
from django.utils import timezone

from freezegun import freeze_time
from germanium.annotations import data_provider
from germanium.test_cases.default import GermaniumTestCase
from germanium.tools import assert_equal
from germanium.tools.models import assert_qs_contains, assert_qs_not_contains
from verification_token.models import VerificationToken

from .base import BaseTestCaseMixin


__all__ = (
   'CleanVerificationTokensCommandTestCase',
)


class CleanVerificationTokensCommandTestCase(BaseTestCaseMixin, GermaniumTestCase):

    @freeze_time(timezone.now())
    @data_provider('create_user')
    def test_clean_verification_tokens_removes_only_non_active_tokens(self, user):
        active_tokens_without_expirations = [VerificationToken.objects.deactivate_and_create(
            obj=user, deactivate_old_tokens=False, expiration_in_minutes=None) for _ in range(10)]
        active_tokens_with_expirations = [VerificationToken.objects.deactivate_and_create(
            obj=user, deactivate_old_tokens=False, expiration_in_minutes=5) for _ in range(10)]
        expired_tokens = [VerificationToken.objects.deactivate_and_create(
            obj=user, deactivate_old_tokens=False, expiration_in_minutes=1) for _ in range(10)]
        deactivated_tokens = [VerificationToken.objects.deactivate_and_create(
            obj=user, deactivate_old_tokens=False, expiration_in_minutes=None) for _ in range(10)]
        VerificationToken.objects.filter(pk__in=[token.pk for token in deactivated_tokens]).update(is_active=False)
        expired_and_deactivated_tokens = [VerificationToken.objects.deactivate_and_create(
            obj=user, deactivate_old_tokens=False, expiration_in_minutes=1) for _ in range(10)]
        VerificationToken.objects.filter(pk__in=[token.pk for token in expired_and_deactivated_tokens]).update(
            is_active=False)

        with freeze_time(timezone.now()+timedelta(minutes=2)):
            call_command('clean_verification_tokens', stdout=StringIO(), stderr=StringIO())

            all_tokens_qs = VerificationToken.objects.all()
            assert_qs_contains(all_tokens_qs, active_tokens_without_expirations)
            assert_qs_contains(all_tokens_qs, active_tokens_with_expirations)
            assert_qs_not_contains(all_tokens_qs, expired_tokens)
            assert_qs_not_contains(all_tokens_qs, deactivated_tokens)
            assert_qs_not_contains(all_tokens_qs, expired_and_deactivated_tokens)



# Generated by Django 2.2.15 on 2020-11-02 19:07

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('verification_token', '0002'),
    ]

    operations = [
        migrations.AlterField(
            model_name='verificationtoken',
            name='object_id',
            field=models.TextField(db_index=True),
        ),
    ]



# Generated by Django 2.2.15 on 2020-11-12 13:38

from django.db import migrations, models
from django.utils.timezone import timedelta


def recompute_value_for_new_field(apps, schema_editor):
    VerificationToken = apps.get_model('verification_token', 'VerificationToken')
    expiring_tokens_qs = VerificationToken.objects.filter(expiration_in_minutes__isnull=False)
    expiration_minutes = expiring_tokens_qs.order_by('expiration_in_minutes').values('expiration_in_minutes').distinct(
        ).values_list('expiration_in_minutes', flat=True)

    for minute_number in expiration_minutes:
        expiring_tokens_qs.filter(expiration_in_minutes=minute_number).update(
            expires_at=models.F('created_at') + timedelta(minutes=minute_number)
        )

    assert VerificationToken.objects.filter(expiration_in_minutes__isnull=False, expires_at__isnull=True).count() == 0


class Migration(migrations.Migration):

    dependencies = [
        ('verification_token', '0003_migration'),
    ]

    operations = [
        migrations.AddField(
            model_name='verificationtoken',
            name='expires_at',
            field=models.DateTimeField(blank=True, default=None, null=True),
        ),
        migrations.RunPython(recompute_value_for_new_field),
    ]



from datetime import timedelta

from django.core.management.base import BaseCommand
from django.db.models import Q
from django.utils import timezone

from verification_token.models import VerificationToken


class Command(BaseCommand):

    def handle(self, **options):
        inactive_and_expired_tokens = VerificationToken.objects.filter(
            Q(is_active=False) | Q(expires_at__isnull=False, expires_at__lt=timezone.now())
        )

        self.stdout.write('Will delete {} inactive or expired verification tokens'.format(
            inactive_and_expired_tokens.count())
        )
        deletion_count = inactive_and_expired_tokens.delete()
        self.stdout.write('Deleted {} inactive or expired verification tokens'.format(deletion_count[0]))
        self.stdout.write('{} verification tokens remain in database'.format(VerificationToken.objects.count()))



# Generated by Django 2.2.15 on 2020-11-12 13:38

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('verification_token', '0004_migration'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='verificationtoken',
            name='expiration_in_minutes',
        ),
    ]
